/*********************************************************************
Author: Antti Hyvarinen <antti.hyvarinen@gmail.com>

OpenSMT2 -- Copyright (C) 2012 - 2014 Antti Hyvarinen
                         2008 - 2012 Roberto Bruttomesso

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*********************************************************************/

#ifndef SSTORE_H
#define SSTORE_H

#include "Sort.h"
#include "SMTConfig.h"
#include "common/StringMap.h"
#include "Alloc.h"

class IdentifierStore
{
  private:
    StrAllocator<IdStr, IdStrRef> isa;
    IdentifierAllocator ia;
  public:
    // Provided for simplicity
    IdRef newIdentifier(const char* name) {
        IdStrRef nr = isa.alloc(name);
        return ia.alloc(nr);
    }
    IdRef newIdentifier(ASTNode& n) {
        if (n.children == NULL) {
            IdStrRef nr = isa.alloc(n.getValue());
            return ia.alloc(nr);
        } else {
            char* s = NULL;
            char* o = s;
            vec<int> nl;
            for (list<ASTNode*>::iterator it = n.children->begin(); it != n.children->end(); it++) {
                o = s;
                asprintf(&s, "%s %s", o, (**it).getValue());
                free(o);
                nl.push(atoi((**it).getValue()));
            }
            o = s;
            asprintf(&s, "%s (%s)", n.getValue(), s);
            free(o);
            IdStrRef nr = isa.alloc(s);
            return ia.alloc(nr, nl);
        }
    }
    const char* getName(IdRef ir) { return isa[ia[ir].getNameRef()].getName(); }
    int* serializeIdentifiers() const;
    void deserializeIdentifiers(const int*);
};

class SStore
{
  private:
    IdentifierStore& is;
    StrAllocator<SStr, SStrRef> ssa;
    SortAllocator sa;
    Map<const char*,SRef,StringHash,Equal<const char*> > sortTable;
    vec<SRef>                                     sorts;

    typedef enum {      // These constants are stored on undo_stack_oper when
        SYMB            // A new symbol is created
      , PARA            // A new parameter
      , CONS            // An undoable cons is done
    } oper_t;

    SMTConfig & config; // Reference to config


  public:

    SStore(SMTConfig & c, IdentifierStore& is_) : is(is_), sa(ssa), config(c) { }

    ~SStore() { }

    //===========================================================================
    // Public APIs for sort construction/destruction

    bool    contains        (const char* s)   const { return sortTable.contains(s); }
    SRef    operator []     (const char* s)   const { return sortTable[s]; }
    bool    contains        (const Sort& s)   const { return sortTable.contains(ssa[s.getNameRef()].getName()); }
    SRef    operator []     (const Sort& s) { return sortTable[ssa[s.getNameRef()].getName()]; }
    Sort*   operator []     (SRef sr)       { return &sa[sr]; }

    char*   buildName       (ASTNode& sn);
    SRef    newSort         (ASTNode& sn);
    SRef    newSort         (IdRef id, vec<SRef>& rest);
    bool    containsSort    (ASTNode& sn)
        { char* name = buildName(sn); bool rval = sortTable.contains(name);
          free(name); return rval; }
    const char* getName     (SRef sr) { return ssa[sa[sr].getNameRef()].getName(); }
    Sort&   getSort         (SRef sr) { return sa[sr]; }

    int*    serializeSorts() const;
    void    deserializeSorts(const int*);
};

#endif
